
/*
	simpledownload.c
*/

#include "simpledownload.h"
#ifdef _MSC_VER
#pragma hdrstop
#endif

#include "util.h"
#include "resource.h"
#include <shellapi.h>

HINSTANCE ghInst ;

static HWND ghWnd ;
static HANDLE g_download_thread ;

static char
 g_url[1024],
 g_dest[1024] ;

static DWORD
 g_app_thread_id ;

enum
{
	error_success = 0,
	error_timeout = 1,
} ;

static DWORD WINAPI download_thread(LPVOID pParam) ;

INLINE BOOL
select_destination(HWND hWnd)
{
    static const char
	 *filter = "All Files (*.*)\0*.*\0" ;	/* delimiter:'\0' */
	OPENFILENAME ofn ;
	char *p ;
	
	p = strrchr(g_url, '/') ;
	if(p)
	{
		p++ ;
		strcpy(g_dest, p) ;
		p = strchr(g_dest, '?') ;
		if(p)
			*p = '\0' ;
	}
	else
		strcpy(g_dest, "untitled") ;

    ofn.lStructSize       = sizeof(ofn) ;
	ofn.hwndOwner		  = hWnd ;
	ofn.lpstrFilter 	  = filter ;
    ofn.lpstrCustomFilter = NULL ;
    ofn.nMaxCustFilter    = 0 ;
    ofn.nFilterIndex      = 0 ;
	ofn.lpstrFile		  = g_dest ;
	ofn.nMaxFile		  = sizeof(g_dest) ;
	ofn.lpstrFileTitle	  = NULL ;
    ofn.nMaxFileTitle     = 0 ;
    /*ofn.lpstrInitialDir   = "d:\\doc\\desktop" ;*/
	ofn.lpstrInitialDir   = "d:\\" ;
	ofn.lpstrTitle		  = load_string(IDS_SAVEAS) ;
	ofn.Flags			  = OFN_HIDEREADONLY |
							/*OFN_OVERWRITEPROMPT |*/
							OFN_ENABLESIZING ;
	ofn.lpstrDefExt 	  = NULL ;
	return GetSaveFileName(&ofn) ;
}

static BOOL CALLBACK
NewDL_DlgProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	switch(uMsg)
	{
	case WM_INITDIALOG:
		CloneWndclass(hWnd) ;
		SetClassLong(
		 hWnd,
		 GCL_HICON,
		 (LONG)LoadIcon(ghInst, (LPCTSTR)IDI_APP)) ;

#ifdef _DEBUG
		SetDlgItemText(
		 hWnd,
		 IDC_URL,
		 "http://freeload.homestead.com/_ksi0701961570623407/floyd24/files/14.mp3") ;
#endif
		return 1 ;

	case WM_COMMAND:
		switch(LOWORD(wParam))
		{
		case IDOK:
			g_url[0] = '\0' ;
			GetDlgItemText(hWnd, IDC_URL, g_url, sizeof(g_url)) ;

			if(g_url[0] == '\0')
			{
				MessageBox(
				 hWnd,
				 load_string(IDS_INVALIDDLURL),
				 g_szAppName,
				 MB_ICONINFORMATION) ;
				return 1 ;	/* not EndDialog */
			}
			
			if(!select_destination(hWnd))	/* if FALSE */
				return 1 ;	/* not EndDialog */

		case IDCANCEL:
			EndDialog(hWnd, LOWORD(wParam)) ;
			break ;

		default:
			return 0 ;
		}
		break ;

	default:
		return 0 ;
	}

	return 1 ;
}

static LRESULT CALLBACK
WndProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	switch(uMsg)
	{
	case WM_SYSCOMMAND:
		switch(LOWORD(wParam))
		{
		case IDM_ALWAYSONTOP:
			{
				BOOL b_always ;
				DWORD style ;

				style = GetWindowLong(hWnd, GWL_EXSTYLE) ;
				b_always = style & WS_EX_TOPMOST ;

				b_always = !b_always ;

				CheckMenuItem(
				 GetSystemMenu(hWnd, FALSE),
				 IDM_ALWAYSONTOP,
				 b_always ? MF_CHECKED : MF_UNCHECKED) ;

				if(b_always)
					style |= WS_EX_TOPMOST ;
				else
					style &= ~WS_EX_TOPMOST ;

				SetWindowLong(hWnd, GWL_EXSTYLE, style) ;

				SetWindowPos(hWnd, NULL, 0, 0, 0, 0,
				 SWP_NOSIZE | SWP_NOMOVE | SWP_NOZORDER | SWP_NOACTIVATE) ;

				style = GetWindowLong(hWnd, GWL_EXSTYLE) ;
			}
			break ;

		case IDM_NEW:
			ShellExecute(
			 NULL,
			 "open",
			 __argv[0],
			 NULL,
			 NULL,
			 SW_SHOWNORMAL) ;
			break ;

		case IDM_START:
			{
				HMENU hMenu ;
				DWORD dwThread ;

				if(g_download_thread)
					break ;

				hMenu = GetSubMenu(GetSystemMenu(ghWnd, FALSE), 0) ;

				EnableMenuItem(hMenu, IDM_START, MF_DISABLED | MF_GRAYED) ;
				EnableMenuItem(hMenu, IDM_PAUSE, MF_ENABLED) ;

				g_download_thread = CreateThread(
				 NULL,
				 0,
				 download_thread,
				 NULL,
				 0,
				 &dwThread) ;
			}
			break ;

		case IDM_PAUSE:
			{
				HMENU hMenu ;

				if(!g_download_thread)
					break ;

				hMenu = GetSubMenu(GetSystemMenu(ghWnd, FALSE), 0) ;

				EnableMenuItem(hMenu, IDM_START, MF_ENABLED) ;
				EnableMenuItem(hMenu, IDM_PAUSE, MF_DISABLED | MF_GRAYED) ;
				
				TerminateThread(g_download_thread, FALSE) ;
				
				set_pause_text(ghWnd, g_dest) ;
				
				g_download_thread = NULL ;
			}
			break ;

		default:
			return DefWindowProc(hWnd, uMsg, wParam, lParam) ;
		}
		break ;

	case WM_DESTROY:
		PostQuitMessage(0) ;
		break ;

	case WM_NCDESTROY:
#ifdef _DEBUG
		OutputDebugString("WM_NCDESTROY\n") ;
#endif
		/* fall through */

	default:
		return DefWindowProc(hWnd, uMsg, wParam, lParam) ;
	}

	return 0 ;
}

INLINE BOOL
create_window()
{
	HMENU hSysMenu ;

	WNDCLASS wc ;
	wc.style			= CS_DBLCLKS ;
	wc.lpfnWndProc		= (WNDPROC)WndProc ;
	wc.cbClsExtra		= 0 ;
	wc.cbWndExtra		= 0 ;
	wc.hInstance		= ghInst ;
	wc.hIcon			= LoadIcon(ghInst, (LPCTSTR)IDI_APP) ;
	wc.hCursor			= LoadCursor(NULL, IDC_ARROW) ;
	wc.hbrBackground	= (HBRUSH)(COLOR_BTNFACE + 1) ;
	wc.lpszMenuName		= NULL ;
	wc.lpszClassName	= "SimpleDownload" ;
	RegisterClass(&wc) ;

	ghWnd = CreateWindowEx(
	 WS_EX_TOPMOST,
	 wc.lpszClassName,
	 NULL,
	 WS_OVERLAPPED | WS_CAPTION | WS_SYSMENU | WS_MINIMIZEBOX | WS_THICKFRAME,
	 CW_USEDEFAULT, CW_USEDEFAULT, 300, 0,
	 0,
	 0,
	 ghInst,
	 NULL) ;
	if(!ghWnd)
		return FALSE ;

	ShowWindow(ghWnd, SW_SHOWDEFAULT) ;

	if(!IsWindowVisible(ghWnd))	/* if SW_SHOWDEFAULT is SW_HIDE */
		return FALSE ;
	
	hSysMenu = GetSystemMenu(ghWnd, FALSE) ;

	InsertMenu(
	 hSysMenu,
	 5,
	 MF_BYPOSITION | MF_CHECKED,
	 IDM_ALWAYSONTOP,
	 load_string(IDM_ALWAYSONTOP_S)) ;

	InsertMenu(
	 hSysMenu,
	 IDM_ALWAYSONTOP,
	 MF_SEPARATOR,
	 0,
	 0) ;

	InsertMenu(
	 hSysMenu,
	 0,
	 MF_BYPOSITION | MF_POPUP,
	 (UINT)LoadMenu(ghInst, (LPCTSTR)IDR_POPUPMENU),
	 "SimpleDownload") ;

	InsertMenu(
	 hSysMenu,
	 1,
	 MF_BYPOSITION | MF_SEPARATOR,
	 0,
	 0) ;

	return TRUE ;
}

static DWORD
search_server(const char *server)
{
	DWORD ip ;

	ip = inet_addr(server) ;
	if(ip == INADDR_NONE)
	{
		struct hostent* pHost = gethostbyname(server) ;
		if(pHost == NULL)
		{
			MessageBox(
			 ghWnd,
			 load_string(IDS_SERVERNOTFOUND),
			 g_szAppName,
			 MB_ICONERROR) ;
			return 0 ;
		}

		ip = *((DWORD*)(pHost->h_addr)) ;
	}

	return ip ;
}

static SOCKET
connect_server(DWORD ip, u_short port)
{
	SOCKADDR_IN addr ;
	int nErr ;
	SOCKET sock ;
	
	sock = socket(PF_INET, SOCK_STREAM, 0) ;
	if(sock == INVALID_SOCKET)
		return 0 ;

	addr.sin_family = AF_INET ;
	addr.sin_port = htons(port) ;
	addr.sin_addr.s_addr = ip ;
	nErr = connect(sock, (const struct sockaddr*)&addr, sizeof(addr)) ;
	if(nErr == SOCKET_ERROR)
		return 0 ;

	return sock ;
}

INLINE void
send_sock(SOCKET sock, const char *s)
{
	send(sock, s, lstrlen(s), 0) ;
#ifdef _DEBUG
	OutputDebugString(s) ;
#endif	/* #ifdef _NDEBUG */
}

static SOCKET
http_connect(const char *url, int start)
{
	SOCKET sock ;
	char
	 *s,
	 *p,
	 *server,
	 *path ;
	int len ;
	char text[1024] ;
	DWORD ip ;

	/* server, path */

	s = (char*)url + 7 ;
	
	p = strchr(s, '/') ;
	if(p)
	{
		len = p - s ;
		path = p ;
	}
	else
	{
		len = strlen(s) ;
		path = s ;
	}

	server = alloca(len + 1) ;
	strncpy(server, s, len) ;
	server[len] = '\0' ;

	/* connect */

	wsprintf(
	 text,
	 load_string(IDS_SEARCHINGSERVER),
	 server) ;
	SetWindowText(ghWnd, text) ;
#ifdef _DEBUG
	OutputDebugString(text) ;
	OutputDebugString("\r\n") ;
#endif /* #ifdef _DEBUG */

	ip = search_server(server) ;

	wsprintf(
	 text,
	 load_string(IDS_CONNECTINGSERVER),
	 server) ;
	SetWindowText(ghWnd, text) ;
#ifdef _DEBUG
	OutputDebugString(text) ;
	OutputDebugString("\r\n") ;
#endif /* #ifdef _DEBUG */

	sock = connect_server(ip, 80) ;
	if(!sock)
		return 0 ;

	/* request 'GET' */

	send_sock(sock, "GET ") ;
	send_sock(sock, path) ;
	send_sock(sock, " HTTP/1.1\r\n") ;

	send_sock(sock, "User-Agent: Mozilla/4.0 (compatible; MSIE 5.0; Windows 2000)\r\n") ;
	
	send_sock(sock, "Accept: */*\r\n") ;
	
	if(start > 0)
	{
		char buf[256] ;
		wsprintf(buf, "Range: bytes=%d-\r\n", start) ;
		send_sock(sock, buf) ;
	}

	send_sock(sock, "Referer: ") ;
	send_sock(sock, url) ;
	send_sock(sock, "\r\n") ;

	send_sock(sock, "Host: ") ;
	send_sock(sock, server) ;
	send_sock(sock, "\r\n") ;

	send_sock(sock, "\r\n") ;	/* end request */

	return sock ;
}

static void
recv_ln(SOCKET sock, char ln[])
{
	char
	 buf[256],
	 *ln_p = ln ;
	
	for(;;)
	{
		int peek_len ;
		const char *p ;

		peek_len = recv(sock, buf, sizeof(buf) - 1, MSG_PEEK) ;
		if(peek_len == 0 || peek_len == SOCKET_ERROR)
			break ;
		buf[peek_len] = '\0' ;

		p = strchr(buf, '\n') ;
		if(p)
		{
			int ln_len, len ;

			ln_len = p - buf + 1 ;	/* 1:'\n' */
			len = recv(sock, buf, ln_len, 0) ;
			if(len == 0 || len == SOCKET_ERROR || len != ln_len)
				break ;
			
			strncpy(ln_p, buf, len) ;
			ln_p += len ;
			
			break ;
		}
		else
		{
			int len = recv(sock, buf, peek_len, 0) ;
			if(len == 0 || len == SOCKET_ERROR)
				break ;
			
			strncpy(ln_p, buf, len) ;
			ln_p += len ;
		}
	}

	while(ln_p[-1] == '\n' || ln_p[-1] == '\r')
		ln_p-- ;

	*ln_p = '\0' ;	/* close string */
}

static int
recv_to_file(SOCKET sock, HANDLE hFile, unsigned content_length)
{
	static const int
	 timeout_sec = 300 ;

	char buf[1024] ;
	int len, n ;
	DWORD dwWritten ;
	
	struct timeval
	 timeout = { timeout_sec, 0 } ;	/* sec, msec */
	
	fd_set fds ;
	FD_ZERO(&fds) ;
	FD_SET(sock, &fds) ;
	
	n = select(0, &fds, NULL, NULL, &timeout) ;
	if(n > 0 && FD_ISSET(sock, &fds))
		;
	else
		return error_timeout ;	/* timeout */

	set_progress_text(ghWnd, content_length, GetFileSize(hFile, NULL), g_dest) ;

	n = sizeof(buf) ;
	if(n + GetFileSize(hFile, NULL) >= content_length)
		n = content_length - GetFileSize(hFile, NULL) ;

	len = recv(sock, buf, sizeof(buf), 0) ;
	if(len == SOCKET_ERROR || len == 0)
		return error_success ;

	WriteFile(
	 hFile,
	 buf,
	 len,
	 &dwWritten,
	 NULL) ;

	return 0 ;
}

static SOCKET
http_start_download(HANDLE hFile, const char *url, int *p_content_length)
{
	SOCKET sock ;
	char ln[1024] ;
	unsigned
	 content_length ;
	int i ;

	for(i = 0 ; i < 6 ; i++)
	{
		sock = http_connect(g_url, GetFileSize(hFile, NULL)) ;
		if(!sock)
			continue ;

		recv_ln(sock, ln) ;	/* expect 'HTTP/1.0 200 OK' */
#ifdef _DEBUG
			OutputDebugString(ln) ;
			OutputDebugString("\r\n") ;
#endif	/* #ifdef _DEBUG */
		
		if(ln[9] == '2')
			break ;
	}

	if(!sock)
		return sock ;
	
	if(ln[9] != '2')
	{
		char buf[1024] ;
		wsprintf(
		 buf,
		 load_string(IDS_INVALIDRESPONSE),
		 ln) ;
		MessageBox(
		 ghWnd,
		 buf,
		 g_szAppName,
		 MB_ICONERROR) ;
		return 0 ;
	}

	content_length = -1 ;
	for(;;)
	{
		recv_ln(sock, ln) ;
		if(ln[0] == '\0')
			break ;

#ifdef _DEBUG
		OutputDebugString(ln) ;
		OutputDebugString("\r\n") ;
#endif	/* #ifdef _DEBUG */

		if(strnicmp(ln, "Content-Length:", 15) == 0)
		{
			const char *p ;
			
			p = ln + 15 ;
			for( ; *p == ' ' || *p == '\t' ; p++)
				;

			content_length = strtol(p, NULL, 10) + GetFileSize(hFile, NULL) ;
		}
	}

	if(p_content_length)
		*p_content_length = content_length ;

	return sock ;
}

static DWORD WINAPI
download_thread(LPVOID pParam)
{
	HANDLE hFile ;
	unsigned content_length ;
	SOCKET sock ;

	/* connecting server */

	if(strnicmp(g_url, "http://", 7) != 0)
	{
		MessageBox(
		 ghWnd,
		 load_string(IDS_INVALIDURL),
		 g_szAppName,
		 MB_ICONERROR) ;
		return FALSE ;
	}

	hFile = CreateFile(
	 g_dest,
	 GENERIC_WRITE,
	 FILE_SHARE_WRITE | FILE_SHARE_READ,
     NULL,
	 OPEN_ALWAYS,
	 FILE_ATTRIBUTE_NORMAL,
	 NULL) ;
	if(hFile == INVALID_HANDLE_VALUE || !hFile)
	{
		char text[1024] ;
		wsprintf(
		 text,
		 load_string(IDS_CREATEFILEFAILED),
		 g_dest) ;
		MessageBox(
		 ghWnd,
		 text,
		 g_szAppName,
		 MB_ICONERROR) ;
		return FALSE ;
	}

	SetFilePointer(hFile, 0, NULL, FILE_END) ;

	sock = http_start_download(hFile, g_url, &content_length) ;

	if(content_length == (unsigned)-1)
	{
		MessageBox(
		 ghWnd,
		 load_string(IDS_INVALIDCONTENTLENGTH),
		 g_szAppName,
		 MB_ICONERROR) ;
		return FALSE ;
	}

	/* downloading */

	while(GetFileSize(hFile, NULL) < content_length)
		recv_to_file(sock, hFile, content_length) ;
	
	CloseHandle(hFile) ;

	/*PostThreadMessage(g_app_thread_id, WM_CLOSE, 0, 0) ;*/
	SetWindowText(ghWnd, "finish") ;

	return TRUE ;
}

int PASCAL
WinMain(HINSTANCE hInst, HINSTANCE hPrevInst, LPSTR pCmdLine, int nShowCmd)
{
	int r ;

	ghInst = hInst ;

	if(!init_winsock())
		return 1 ;

	r = DialogBox(
	 hInst,
	 (LPCTSTR)IDD_NEWDL,
	 0,
	 NewDL_DlgProc) ;
	if(!r || r == IDCANCEL)
		return 2 ;

	if(!create_window())
		return 3 ;

	g_app_thread_id = GetCurrentThreadId() ;

	SendMessage(
	 ghWnd, WM_SYSCOMMAND, MAKELONG(IDM_START, 0), 0) ;

	if(ghWnd
	&& IsWindowVisible(ghWnd))
	{
		MSG msg ;
		for(;;)
		{
			if(PeekMessage(&msg, NULL, 0, 0, PM_NOREMOVE))
			{
				if(!GetMessage(&msg, NULL, 0, 0))
					return msg.wParam ;
           
				TranslateMessage(&msg) ;
				DispatchMessage(&msg) ;
			}
		}
	}

	return 0 ;
}
